from grpc import ChannelConnectivity, StatusCode, RpcError, RpcContext,\
                ChannelCredentials, CallCredentials, ssl_channel_credentials,\
                metadata_call_credentials,access_token_call_credentials,\
                composite_call_credentials,composite_channel_credentials
import asyncio as _asyncio
import functools as _functools
from grpc import insecure_channel as _insecure_channel
from grpc import secure_channel as _secure_channel
from .channel import Channel, insecure_channel, secure_channel, channel_ready_future




'''
Created on 2017/8/14

:author: hubo
'''

import asyncio
import functools
import queue
from asyncio import CancelledError

def wrap_callback(callback, loop):
    @functools.wraps(callback)
    def _callback(*args, **kwargs):
        if not loop.is_closed():
            loop.call_soon_threadsafe(functools.partial(callback, *args, **kwargs))
    return _callback


def wrap_active_test(func, test, loop, executor=None):
    @functools.wraps(func)
    async def _func(*args, **kwargs):
        if test():
            return await loop.run_in_executor(executor, functools.partial(func, *args, **kwargs))
        else:
            return func(*args, **kwargs)
    return _func


def wrap_future(grpc_fut, loop):
    fut_ = loop.create_future()
    def _set_state(grpc_fut, fut_):
        assert grpc_fut.done()
        if fut_.cancelled():
            return
        assert not fut_.done()
        if grpc_fut.cancelled():
            fut_.cancel()
        else:
            exception = grpc_fut.exception()
            if exception is not None:
                fut_.set_exception(exception)
            else:
                result = grpc_fut.result()
                fut_.set_result(result)

    def _call_check_cancel(fut_):
        if fut_.cancelled():
            grpc_fut.cancel()

    def _call_set_state(grpc_fut):
        if not loop.is_closed():
            loop.call_soon_threadsafe(_set_state, grpc_fut, fut_)

    fut_.add_done_callback(_call_check_cancel)
    grpc_fut.add_done_callback(_call_set_state)
    return fut_


def copy_members(source, dest, member_list, wrapper=None):
    for m in member_list:
        f = getattr(source, m, None)
        if f is None:
            continue
        if wrapper is not None:
            f = wrapper(f)
        setattr(dest, m, f)


def wrap_future_call(grpc_fut, loop, executor=None):
    fut_ = wrap_future(grpc_fut, loop)
    # Copy extra members
    copy_members(grpc_fut, fut_,
                 ['is_active',
                  'time_remaining'])
    @functools.wraps(grpc_fut.add_callback)
    def _add_callback(callback):
        grpc_fut.add_callback(wrap_callback(callback, loop))
    fut_.add_callback = _add_callback
    copy_members(grpc_fut, fut_,
                 ['initial_metadata',
                  'trailing_metadata',
                  'code',
                  'details'],
                 functools.partial(wrap_active_test, test=grpc_fut.is_active, loop=loop, executor=executor))
    return fut_


class WrappedIterator(object):
    """
    Wrap an grpc_iterator to an async iterator
    """
    def __init__(self, grpc_iterator, loop, executor=None, stream_executor=None):
        self._iterator = grpc_iterator
        self._loop = loop
        self._executor = executor
        if stream_executor is None:
            self._shared_executor = True
            self._stream_executor = executor
        else:
            self._shared_executor = False
            self._stream_executor = stream_executor
        self._next_future = None
        copy_members(grpc_iterator, self,
                     ['is_active',
                      'time_remaining',
                      'cancel'])
        @functools.wraps(grpc_iterator.add_callback)
        def _add_callback(callback):
            grpc_iterator.add_callback(wrap_callback(callback, loop))
        self.add_callback = _add_callback
        copy_members(grpc_iterator, self,
                     ['initial_metadata',
                      'trailing_metadata',
                      'code',
                      'details'],
                     functools.partial(wrap_active_test, test=grpc_iterator.is_active, loop=loop, executor=executor))        
        
    def __aiter__(self):
        return self
    
    def _next(self):
        if self._iterator is None:
            raise StopAsyncIteration
        try:
            return next(self._iterator)
        except StopIteration:
            raise StopAsyncIteration
        except Exception:
            raise
    
    async def __anext__(self):
        if self._next_future is None:
            if self._iterator is None:
                raise StopAsyncIteration
            self._next_future = self._loop.run_in_executor(self._stream_executor, self._next)
        try:
            return await asyncio.shield(self._next_future, loop=self._loop)
        finally:
            if self._next_future and self._next_future.done():
                self._next_future = None
    
    def __del__(self):
        if self._iterator is not None:
            self.cancel()
            self._iterator = None
        if self._next_future is not None:
            if not self._loop.is_closed():
                self._loop.call_soon_threadsafe(lambda f=self._next_future: f.cancel())
            self._next_future = None
        if not self._shared_executor and self._stream_executor is not None:
            self._stream_executor.shutdown()
            self._stream_executor = None

    async def aclose(self):
        self.__del__()


class IteratorScope(object):
    def __init__(self, _iter):
        self._iter = _iter

    async def __aenter__(self):
        return self._iter
    
    async def __aexit__(self, exc_val, exc_typ, exc_tb):
        await self._iter.aclose()


class WrappedAsyncIterator(object):
    """
    Wrap an async iterator to an iterator for grpc input
    """
    def __init__(self, async_iter, loop):
        self._async_iter = async_iter
        self._loop = loop
        self._q = queue.Queue()
        self._stop_future = loop.create_future()
        self._next_future = None
        self._closed = False

    def __iter__(self):
        return self

    async def _next(self):
        if self._async_iter is None:
            # An edge condition
            self._q.put((None, True))
            return
        if self._next_future is None:
            self._next_future = asyncio.ensure_future(self._async_iter.__anext__(), loop=self._loop)
        try:
            done, _ = await asyncio.wait([self._stop_future, self._next_future], loop=self._loop,
                                   return_when=asyncio.FIRST_COMPLETED)
            if self._stop_future in done:
                self._q.put((await self._stop_future, True))
                self._next_future.cancel()
                try:
                    await self._next_future
                except CancelledError:
                    pass
                finally:
                    self._next_future = None
            else:
                nf = self._next_future
                self._next_future = None
                self._q.put((await nf, False))
        except StopAsyncIteration:
            self._q.put((None, True))
        except Exception as exc:
            self._q.put((exc, True))
        
    def __next__(self):
        if self._async_iter is None:
            raise StopIteration
        try:
            r, is_exc = self._q.get_nowait()
        except queue.Empty:
            if not self._loop.is_closed():
                self._loop.call_soon_threadsafe(functools.partial(asyncio.ensure_future, self._next(), loop=self._loop))
            r, is_exc = self._q.get()
        if is_exc:
            if r is None:
                self._async_iter = None
                raise StopIteration
            else:
                raise r
        else:
            return r
    
    def close(self):
        if self._async_iter is not None:
            async def async_close():
                if not self._stop_future.done():
                    self._stop_future.set_result(None)
                await self._async_iter.aclose()
            try:
                if not self._loop.is_closed():
                    self._loop.call_soon_threadsafe(functools.partial(asyncio.ensure_future, async_close(), loop=self._loop))
            finally:
                # Ensure __next__ ends
                self._q.put((None, True))
                self._async_iter = None

    def cancel(self, exception=True):
        if exception:
            exc = CancelledError()
        else:
            exc = None
        def _set_result():
            if not self._stop_future.done():
                self._stop_future.set_result(exc)
        # Ensure __next__ ends. Sometimes the loop is already closing, so the exit result may not be written
        # to the queue
        self._q.put((exc, True))
        if not self._loop.is_closed():
            self._loop.call_soon_threadsafe(_set_result)



'''
Created on 2017/8/11

:author: hubo
'''
import grpc as _grpc
import aiogrpc.utils as _utils
import asyncio as _asyncio
from concurrent.futures import ThreadPoolExecutor as _ThreadPoolExecutor
import functools


class _UnaryUnaryMultiCallable(object):
    """Affords invoking a unary-unary RPC from client-side."""
    def __init__(self, _inner, loop, executor=None):
        self._inner = _inner
        self._loop = loop
        self._executor = executor
    
    async def __call__(self, request, timeout=None, metadata=None, credentials=None):
        """Synchronously invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      timeout: An optional duration of time in seconds to allow for the RPC.
      metadata: Optional :term:`metadata` to be transmitted to the
        service-side of the RPC.
      credentials: An optional CallCredentials for the RPC.

    Returns:
      The response value for the RPC.

    Raises:
      RpcError: Indicating that the RPC terminated with non-OK status. The
        raised RpcError will also be a Call for the RPC affording the RPC's
        metadata, status code, and details.
    """
        fut = self.future(request, timeout, metadata, credentials)
        try:
            return await fut
        finally:
            if not fut.done():
                fut.cancel()

    async def with_call(self, request, timeout=None, metadata=None, credentials=None):
        """Synchronously invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      timeout: An optional durating of time in seconds to allow for the RPC.
      metadata: Optional :term:`metadata` to be transmitted to the
        service-side of the RPC.
      credentials: An optional CallCredentials for the RPC.

    Returns:
      The response value for the RPC and a Call value for the RPC.

    Raises:
      RpcError: Indicating that the RPC terminated with non-OK status. The
        raised RpcError will also be a Call for the RPC affording the RPC's
        metadata, status code, and details.
    """
        fut = self.future(request, timeout, metadata, credentials)
        try:
            result = await fut
            return (result, fut)
        finally:
            if not fut.done():
                fut.cancel()


    def future(self, request, timeout=None, metadata=None, credentials=None):
        """Asynchronously invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      timeout: An optional duration of time in seconds to allow for the RPC.
      metadata: Optional :term:`metadata` to be transmitted to the
        service-side of the RPC.
      credentials: An optional CallCredentials for the RPC.

    Returns:
        An object that is both a Call for the RPC and a Future. In the event of
        RPC completion, the return Call-Future's result value will be the
        response message of the RPC. Should the event terminate with non-OK
        status, the returned Call-Future's exception value will be an RpcError.
    """
        return _utils.wrap_future_call(self._inner.future(request, timeout, metadata, credentials),
                                       self._loop, self._executor)


class _UnaryStreamMultiCallable(object):
    """Affords invoking a unary-stream RPC from client-side."""
    def __init__(self, _inner, loop, executor=None, standalone_pool=False):
        self._inner = _inner
        self._loop = loop
        self._executor = executor
        self._standalone_pool = standalone_pool
    
    def __call__(self, request, timeout=None, metadata=None, credentials=None,
                    *, standalone_pool = None):
        """Invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      timeout: An optional duration of time in seconds to allow for the RPC.
               If None, the timeout is considered infinite.
      metadata: An optional :term:`metadata` to be transmitted to the
        service-side of the RPC.
      credentials: An optional CallCredentials for the RPC.

    Returns:
        An object that is both a Call for the RPC and an ASYNC iterator of response
        values. Drawing response values from the returned Call-iterator may
        raise RpcError indicating termination of the RPC with non-OK status.
    """
        if standalone_pool is None:
            standalone_pool = self._standalone_pool
        if standalone_pool:
            stream_executor = _ThreadPoolExecutor(1)
        else:
            stream_executor = None
        return _utils.WrappedIterator(self._inner(request, timeout, metadata, credentials),
                                      self._loop,
                                      self._executor,
                                      stream_executor)
    
    def with_scope(self, request, timeout=None, metadata=None, credentials=None,
                        *, standalone_pool = None):
        """
        Return an ASYNC context manager to ensure the call is closed outside the scope::
        
            async with mystub.mymethod.with_scope(...) as iter:
                async for i in iter:
                    ...
        """
        return _utils.IteratorScope(self(request, timeout, metadata, credentials, standalone_pool=standalone_pool))


class _StreamUnaryMultiCallable(object):
    """Affords invoking a stream-unary RPC from client-side."""
    def __init__(self, _inner, loop, executor=None):
        self._inner = _inner
        self._loop = loop
        self._executor = executor
        
    async def __call__(self,
                 request_iterator,
                 timeout=None,
                 metadata=None,
                 credentials=None):
        """Synchronously invokes the underlying RPC.

    Args:
      request_iterator: An ASYNC iterator that yields request values for the RPC.
      timeout: An optional duration of time in seconds to allow for the RPC.
               If None, the timeout is considered infinite.
      metadata: Optional :term:`metadata` to be transmitted to the
        service-side of the RPC.
      credentials: An optional CallCredentials for the RPC.

    Returns:
      The response value for the RPC.

    Raises:
      RpcError: Indicating that the RPC terminated with non-OK status. The
        raised RpcError will also implement grpc.Call, affording methods
        such as metadata, code, and details.
    """
        fut = self.future(request_iterator, timeout, metadata, credentials)
        try:
            return await fut
        finally:
            if not fut.done():
                fut.cancel()

    async def with_call(self,
                  request_iterator,
                  timeout=None,
                  metadata=None,
                  credentials=None):
        """Synchronously invokes the underlying RPC on the client.

    Args:
      request_iterator: An ASYNC iterator that yields request values for the RPC.
      timeout: An optional duration of time in seconds to allow for the RPC.
               If None, the timeout is considered infinite.
      metadata: Optional :term:`metadata` to be transmitted to the
        service-side of the RPC.
      credentials: An optional CallCredentials for the RPC.

    Returns:
      The response value for the RPC and a Call object for the RPC.

    Raises:
      RpcError: Indicating that the RPC terminated with non-OK status. The
        raised RpcError will also be a Call for the RPC affording the RPC's
        metadata, status code, and details.
    """
        fut = self.future(request_iterator, timeout, metadata, credentials)
        try:
            result = await fut
            return (result, fut)
        finally:
            if not fut.done():
                fut.cancel()

    def future(self,
               request_iterator,
               timeout=None,
               metadata=None,
               credentials=None):
        """Asynchronously invokes the underlying RPC on the client.

    Args:
      request_iterator: An ASYNC iterator that yields request values for the RPC.
      timeout: An optional duration of time in seconds to allow for the RPC.
               If None, the timeout is considered infinite.
      metadata: Optional :term:`metadata` to be transmitted to the
        service-side of the RPC.
      credentials: An optional CallCredentials for the RPC.

    Returns:
        An object that is both a Call for the RPC and a Future. In the event of
        RPC completion, the return Call-Future's result value will be the
        response message of the RPC. Should the event terminate with non-OK
        status, the returned Call-Future's exception value will be an RpcError.
    """
        return _utils.wrap_future_call(
                    self._inner.future(
                        _utils.WrappedAsyncIterator(request_iterator, self._loop),
                        timeout,
                        metadata,
                        credentials
                    ),
                    self._loop,
                    self._executor)


class _StreamStreamMultiCallable(object):
    """Affords invoking a stream-stream RPC on client-side."""

    def __init__(self, _inner, loop, executor=None, standalone_pool=False):
        self._inner = _inner
        self._loop = loop
        self._executor = executor
        self._standalone_pool = standalone_pool

    def __call__(self,
                 request_iterator,
                 timeout=None,
                 metadata=None,
                 credentials=None,
                 *, standalone_pool = None):
        """Invokes the underlying RPC on the client.

    Args:
      request_iterator: An iterator that yields request values for the RPC.
      timeout: An optional duration of time in seconds to allow for the RPC.
               if not specified the timeout is considered infinite.
      metadata: Optional :term:`metadata` to be transmitted to the
        service-side of the RPC.
      credentials: An optional CallCredentials for the RPC.

    Returns:
        An object that is both a Call for the RPC and an iterator of response
        values. Drawing response values from the returned Call-iterator may
        raise RpcError indicating termination of the RPC with non-OK status.
    """
        if standalone_pool is None:
            standalone_pool = self._standalone_pool
        if standalone_pool:
            stream_executor = _ThreadPoolExecutor(1)
        else:
            stream_executor = None
        input_iterator = _utils.WrappedAsyncIterator(request_iterator,
                                                    self._loop)
        grpc_iterator = self._inner(
                        input_iterator,
                        timeout,
                        metadata,
                        credentials
                    )
        r = _utils.WrappedIterator(
                    grpc_iterator,
                    self._loop,
                    self._executor,
                    stream_executor)
        # Make sure the input iterator exits, or the thread may block indefinitely
        def _end_callback():
            input_iterator.cancel(False)
        grpc_iterator.add_callback(_end_callback)
        return r


    def with_scope(self, request_iterator, timeout=None, metadata=None, credentials=None,
                   *, standalone_pool = None):
        """
        Return an ASYNC context manager to ensure the call is closed outside the scope::
        
            async with mystub.mymethod.with_scope(...) as iter:
                async for i in iter:
                    ...
        """
        return _utils.IteratorScope(self(request_iterator, timeout, metadata, credentials,
                                         standalone_pool=standalone_pool))


class Channel(_grpc.Channel):
    """An asyncio proxy of grpc.Channel."""

    def __init__(self, _channel, loop=None, executor=None, standalone_pool_for_streaming=False):
        """Constructor.

        Args:
          _channel: wrapped grpc.Channel
          loop: asyncio event loop
          executor: a thread pool, or None to use the default pool of the loop
          standalone_pool_for_streaming: create a new thread pool (with 1 thread) for each streaming
                                         method
        """
        self._channel = _channel
        if loop is None:
            loop = _asyncio.get_event_loop()
        self._loop = loop
        self._executor = executor
        self._standalone_pool = standalone_pool_for_streaming
        self._subscribe_map = {}

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    def subscribe(self, callback, try_to_connect=None):
        if callback in self._subscribe_map:
            wrapped_callback = self._subscribe_map[callback]
        else:
            wrapped_callback = _utils.wrap_callback(callback, self._loop)
        self._channel.subscribe(wrapped_callback, try_to_connect)
        self._subscribe_map[callback] = wrapped_callback

    def unsubscribe(self, callback):
        if callback in self._subscribe_map:
            self._channel.unsubscribe(self._subscribe_map[callback])
            del self._subscribe_map[callback]

    def unary_unary(self,
                    method,
                    request_serializer=None,
                    response_deserializer=None):
        return _UnaryUnaryMultiCallable(
                    self._channel.unary_unary(method, request_serializer, response_deserializer),
                    self._loop,
                    self._executor)

    def unary_stream(self,
                     method,
                     request_serializer=None,
                     response_deserializer=None):
        return _UnaryStreamMultiCallable(
            self._channel.unary_stream(method, request_serializer, response_deserializer),
            self._loop,
            self._executor,
            self._standalone_pool)

    def stream_unary(self,
                     method,
                     request_serializer=None,
                     response_deserializer=None):
        return _StreamUnaryMultiCallable(
            self._channel.stream_unary(method, request_serializer, response_deserializer),
            self._loop,
            self._executor)

    def stream_stream(self,
                      method,
                      request_serializer=None,
                      response_deserializer=None):
        return _StreamStreamMultiCallable(
            self._channel.stream_stream(method, request_serializer, response_deserializer),
            self._loop,
            self._executor,
            self._standalone_pool)

    def close(self):
        """
        This method returns a future, so you should use `await channel.close()`, but call
        `channel.close()` also close it (but did not wait for the closing)
        """
        return self._loop.run_in_executor(self._executor, self._channel.close)


def channel_ready_future(channel):
    """Creates a Future that tracks when a Channel is ready.

  Cancelling the Future does not affect the channel's state machine.
  It merely decouples the Future from channel state machine.

  Args:
    channel: A Channel object.

  Returns:
    A Future object that matures when the channel connectivity is
    ChannelConnectivity.READY.
  """
    fut = channel._loop.create_future()
    def _set_result(state):
        if not fut.done() and state is _grpc.ChannelConnectivity.READY:
            fut.set_result(None)
    fut.add_done_callback(lambda f: channel.unsubscribe(_set_result))
    channel.subscribe(_set_result, try_to_connect=True)
    return fut


def insecure_channel(target, options=None, *, loop=None, executor=None,
                    standalone_pool_for_streaming=False):
    """Creates an insecure Channel to a server.

  Args:
    target: The server address
    options: An optional list of key-value pairs (channel args in gRPC runtime)
    to configure the channel.

  Returns:
    A Channel object.
  """
    return Channel(_grpc.insecure_channel(target, options), loop, executor, standalone_pool_for_streaming)


def secure_channel(target, credentials, options=None, *, loop=None, executor=None,
                   standalone_pool_for_streaming=False):
    """Creates a secure Channel to a server.

  Args:
    target: The server address.
    credentials: A ChannelCredentials instance.
    options: An optional list of key-value pairs (channel args in gRPC runtime)
    to configure the channel.

  Returns:
    A Channel object.
  """
    return Channel(_grpc.secure_channel(target, credentials, options),
                   loop, executor, standalone_pool_for_streaming)



#!/usr/bin/env python

from setuptools import setup

version = "1.7"

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = ""


setup(
    name="aiogrpc",
    version=version,
    author="Hu Bo",
    author_email="hubo1016@126.com",
    long_description=long_description,
    description="asyncio wrapper for grpc.io",
    license="Apache",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Framework :: AsyncIO',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Framework :: AsyncIO',
    ],
    url="https://github.com/hubo1016/aiogrpc",
    platforms=['any'],
    packages=[
        'aiogrpc',
    ],
    python_requires='>=3.6',
    install_requires=[
        'grpcio>=1.12.0'
    ]
)
